{ import: token }

LineScanner : Object ( input line lineNumber )

LineScanner on: aStringOrStream
[
    self := self new initialize.
    input := aStringOrStream readStream.
]

LineScanner initialize
[
    line := WriteStream on: (String new: 256).
    lineNumber := 0.
]

" 5.1.1.2  Translation phases

  1. Physical source file multibyte characters are mapped, in an implementation-
     defined manner, to the source character set (introducing new-line characters for
     end-of-line indicators) if necessary.  Trigraph sequences are replaced by 
     corresponding single-character internal representations. "

LineScanner discardEOL
[
    | c |
    lineNumber := lineNumber + 1.
    c := input next.
    c == $\n
	ifTrue:  [[input peek == $\r] whileTrue: [input next]]
	ifFalse:
	   [c == $\n
		ifTrue:  [[input peek == $\r] whileTrue: [input next]]
		ifFalse: [self error: 'no EOL']]
]

LineScanner retainEOL
[
    line nextPut: $\n; reset.
    self discardEOL.
]

" 2. Each instance of a backslash character (\) immediately followed by a new-line 
     character is deleted, splicing physical source lines to form logical source lines. 
     Only the last backslash on any physical source line shall be eligible for being part 
     of such a splice.  A source file that is not empty shall end in a new-line character, 
     which shall not be immediately preceded by a backslash character before any such 
     splicing takes place. "

LineScanner refill
[
    | c |
    line reset.
    input atEnd ifTrue: [^line reset].
    [(c := input peek) notNil]
	whileTrue:
	   [c == $\\
		ifTrue:
		   [input next.
		    (input peek == $\n or: [input peek == $\r])
			ifTrue:  [self discardEOL]
			ifFalse: [line nextPut: c]]
		ifFalse:
		   [(c == $\n or: [c == $\r])
			ifTrue:  [^self retainEOL]
			ifFalse: [line nextPut: input next]]].
    line reset.
]

LineScanner next
[
    line atEnd ifTrue: [self refill].
    ^line atEnd ifFalse: [line next].
]

LineScanner peek
[
    line atEnd ifTrue: [self refill].
    ^line atEnd ifFalse: [line peek].
]

LineScanner match: aCollection
[
    0 to: aCollection size - 1 do: [:i | (line peek: i) == (aCollection at: i) ifFalse: [^false]].
    ^true
]

LineScanner skip: aCollection
[
    (self match: aCollection) ifFalse: [^false].
    line skip: aCollection size.
    ^true
]

"----------------------------------------------------------------"

C99CharacterTable := [
    (Array new: 256 withAll: #xIllegal)
	atAll: '!'	 put: #xPling;
	atAll: '#'	 put: #xHash;
	atAll: '%'	 put: #xPercent;
	atAll: '&'	 put: #xAmpersand;
	atAll: '('	 put: #xLparen;
	atAll: ')'	 put: #xRparen;
	atAll: '*'	 put: #xStar;
	atAll: '+'	 put: #xPlus;
	atAll: ','	 put: #xComma;
	atAll: '-'	 put: #xMinus;
	atAll: '.'	 put: #xDot;
	atAll: '/'	 put: #xSlash;
	atAll: ':'	 put: #xColon;
	atAll: ';'	 put: #xSemicolon;
	atAll: '<'	 put: #xLangle;
	atAll: '='	 put: #xEqual;
	atAll: '>'	 put: #xRangle;
	atAll: '?'	 put: #xQuery;
	atAll: '['	 put: #xLbracket;
	atAll: ']'	 put: #xRbracket;
	atAll: '{'	 put: #xLbrace;
	atAll: '|'	 put: #xBar;
	atAll: '}'	 put: #xRbrace;
	atAll: '~'	 put: #xTilde;
	atAll: '^'	 put: #xHat;
	atAll: '_'	 put: #xLetter;
	from:  $a to: $z put: #xLetter;
	from:  $A to: $Z put: #xLetter;
	from:  $0 to: $9 put: #xDigit;
	from:   0 to: 32 put: #xSpace;
	at:    $"	 put: #xString;
	at:    $'	 put: #xChar;
	at:    $$	 put: #xOther;
	at:    $@	 put: #xOther;
	at:    $\n	 put: #xEOL;
	yourself
]

C99DirectiveTable := [
    IdentityDictionary new
	at: #if		 put: C99DirectiveIf;
	at: #ifdef	 put: C99DirectiveIfdef;
	at: #ifndef	 put: C99DirectiveIfndef;
	at: #elif	 put: C99DirectiveElif;
	at: #else	 put: C99DirectiveElse;
	at: #endif	 put: C99DirectiveEndif;
	at: #include	 put: C99DirectiveInclude;
	at: #define	 put: C99DirectiveDefine;
	at: #undef	 put: C99DirectiveUndef;
	at: #line	 put: C99DirectiveLine;
	at: #error	 put: C99DirectiveError;
	at: #pragma	 put: C99DirectivePragma;
	yourself
]

C99StringTable := [
    | table |
    table := Array new: 256.
    0 to: 255 do: [:c | table at: c put: (String with: c)].
    table
]

C99Scanner : LineScanner ( characterTable thisChar text bol inDirective activeGroup tokenFileName tokenLineNumber )

C99Scanner on: fileOrString
[
    self := super on: fileOrString.
    tokenFileName := fileOrString isFile ifTrue:  [fileOrString name] ifFalse: ['<none>']
]

C99Scanner initialize
[
    super initialize.
    characterTable := C99CharacterTable.
    text := WriteStream on: (String new: 32).
    self beginningOfLine.
    bol := true.
    inDirective := false.
    tokenLineNumber := 0.
]

C99Scanner refill
[
    super refill.
    tokenLineNumber := lineNumber.
]

" 6.4  Lexical elements

  The categories of preprocessing tokens are: header names, identifiers,
  preprocessing numbers, character constants, string literals, punctuators, and
  single non-white-space characters that do not lexically match the other
  preprocessing token categories.  ... "

C99Scanner identifier			[ bol := false.  ^C99Identifier withText: text contents asSymbol    file: tokenFileName line: tokenLineNumber 		]
C99Scanner number			[ bol := false.  ^C99Number	withText: text contents		    file: tokenFileName line: tokenLineNumber  	    	]
C99Scanner blank			[		 ^C99Blank	withText: text contents		    file: tokenFileName line: tokenLineNumber  	   	]
C99Scanner blank: aNote			[		 ^C99Blank	withText: text contents		    file: tokenFileName line: tokenLineNumber note: aNote	]
C99Scanner punctuator: tokText op: type	[ bol := false.  ^type		withText: tokText		    file: tokenFileName line: tokenLineNumber 		]
C99Scanner other: char			[ bol := false.  ^C99Other	withText: (C99StringTable at: char) file: tokenFileName line: tokenLineNumber 		]

C99Scanner scanText: category until: terminator	"character and string"
[
    | start c |
    bol := false.
    start := line position.
    text nextPut: self next.
    [(c := self next) notNil and: [c ~~ $\n]]
	whileTrue:
	   [c == terminator ifTrue: [^category withText: (text nextPut: c; contents) file: tokenFileName line: tokenLineNumber].
	    text nextPut: c.
	    (c == $\\ and: [(c := self peek) notNil and: [c ~~ $\n]]) ifTrue: [text nextPut: self next]].
    "...  If a ' or a '' character matches the last category, the behavior is undefined. "
    line position: start.
    ^self other: self next
]

" 5.1.1.2  Translation phases

  3. The source file is decomposed into preprocessing tokens and sequences of 
     white-space characters (including comments).  A source file shall not end in a 
     partial preprocessing token or in a partial comment.  Each comment is replaced by 
     one space character.  New-line characters are retained.  Whether each nonempty 
     sequence of white-space characters other than new-line is retained or replaced by 
     one space character is implementation-defined. "

C99Scanner nextToken
[
    tokenLineNumber := lineNumber.
    (thisChar := self peek) notNil ifTrue: [^self perform: (characterTable at: thisChar)].
    activeGroup notNil ifTrue: [activeGroup error: 'unterminated #if'].
    ^nil
]

C99Scanner scan
[
    | tokens token |
    tokens := C99TokenList new.
    [token := self nextToken] whileTrue: [tokens addLast: token].
    ^tokens
]

" white-space "

C99Scanner xSpace
[
    | c |
    text reset; nextPut: self next.
    [(c := self peek) notNil and: [(characterTable at: c) == #xSpace]] whileTrue: [text nextPut: self next].
    ^self blank
]

C99Scanner xEOL
[
    self beginningOfLine.
    text reset; nextPut: self next.
    ^self blank: #eol
]

C99Scanner beginningOfLine	[ bol := true ]

C99Scanner multiComment		"sent from /"
[
    text reset; nextPut: self next; nextPut: self next.
    [self peek; match: '*/'] whileFalse: [text nextPut: self next].
    text nextPut: self next; nextPut: self next.
    ^self blank: #comment
]

C99Scanner uniComment		"sent from /"
[
    | c |
    text reset; nextPut: self next; nextPut: self next.
    [(c := self peek) notNil and: [c ~~ $\n]] whileTrue: [text nextPut: self next].
    text nextPut: self next.
    ^self blank: #eol
]

" single non-white-space "

C99Scanner xOther	[ ^self other: self next ]
C99Scanner xIllegal	[ ^self other: self next ]

" character constants and string literals "

C99Scanner xChar	[ text reset.  ^self scanText: C99Character until: $']
C99Scanner xString	[ text reset.  ^self scanText: C99String    until: $"]

" identifiers "

C99Scanner xLetter
[
    | c |
    text reset.
    thisChar == $L
	ifTrue:
	   [(line peek: 1) == $" ifTrue: [text nextPut: self next.  ^self scanText: C99String    until: $"].
	    (line peek: 1) == $' ifTrue: [text nextPut: self next.  ^self scanText: C99Character until: $']].
    [(c := self peek) notNil and: [(characterTable at: c) == #xLetter or: [(characterTable at: c) == #xDigit]]]
	whileTrue: [text nextPut: self next].
    ^self identifier
]

" pp-number :
	. digit "

C99Scanner xDot
[
    | d |
    text reset.
    ((d := line peek: 1) notNil and: [(characterTable at: d) == #xDigit]) ifTrue: [^self xDigit].
    (self skip: '...') ifTrue: [^self punctuator: '...' op: C99Ellipsis].
    ^self next; punctuator: '.' op: C99Member
]

" pp-number :
	digit "

C99Scanner xDigit
[
    text reset.
    [self ppNumber] whileTrue.
    ^self number.
]

" pp-number :
	pp-number digit
	pp-number identifier-nondigit
	pp-number e sign
	pp-number E sign
	pp-number p sign
	pp-number P sign
	pp-number . "

C99Scanner ppNumber
[
    | c d |
    (c := self peek) isNil ifTrue: [^false].
    (characterTable at: c) == #xDigit  ifTrue: [text nextPut: self next.  ^true].
    (characterTable at: c) == #xLetter ifTrue: [text nextPut: self next.  ^true].
    c == $. ifTrue: [text nextPut: self next.  ^true].
    (d := line peek: 1) isNil ifTrue: [^false].
    ((c == $e or: [c == $E or: [c == $p or: [c == $P]]]) and: [d == $+ or: [d == $-]])
	ifTrue: [text nextPut: self next; nextPut: self next.  ^true].
    ^false
]

" punctuators "

C99Scanner xLparen				[ ^self next; 			     punctuator: '('   op: C99OpenParen ]
C99Scanner xRparen				[ ^self next; 			     punctuator: ')'   op: C99CloseParen ]
C99Scanner xComma				[ ^self next; 			     punctuator: ','   op: C99Comma ]
C99Scanner xSemicolon				[ ^self next; 			     punctuator: ';'   op: C99Semicolon ]
C99Scanner xQuery				[ ^self next; 			     punctuator: '?'   op: C99Query ]
C99Scanner xLbracket				[ ^self next; 			     punctuator: '['   op: C99OpenBracket ]
C99Scanner xRbracket				[ ^self next; 			     punctuator: ']'   op: C99CloseBracket ]
C99Scanner xLbrace				[ ^self next; 			     punctuator: '{'   op: C99OpenBrace ]
C99Scanner xRbrace				[ ^self next; 			     punctuator: '}'   op: C99CloseBrace ]
C99Scanner xTilde				[ ^self next; 			     punctuator: '~'   op: C99Complement ]
C99Scanner xSlash	" /* // /= / "		[ (self match: '/*' ) ifTrue: [^self multiComment].
						  (self match: '//' ) ifTrue: [^self uniComment].
						  (self skip:  '/=' ) ifTrue: [^self punctuator: '/='  op: C99DivideInto].
						  ^self next;			     punctuator: '/'   op: C99Divide]
C99Scanner xLangle	" <<= << <= <% <: < "	[ (self skip: '<<=' ) ifTrue: [^self punctuator: '<<=' op: C99LeftShiftInto].
						  (self skip: '<<'  ) ifTrue: [^self punctuator: '<<'  op: C99LeftShift].
						  (self skip: '<='  ) ifTrue: [^self punctuator: '<='  op: C99LessEqual].
						  (self skip: '<%'  ) ifTrue: [^self punctuator: '{'   op: C99OpenBrace].
						  (self skip: '<:'  ) ifTrue: [^self punctuator: '['   op: C99OpenBracket].
						  ^self next;			     punctuator: '<'   op: C99Less]
C99Scanner xRangle	" >>= >> >= > "		[ (self skip: '>>=' ) ifTrue: [^self punctuator: '>>=' op: C99RightShiftInto].
						  (self skip: '>>'  ) ifTrue: [^self punctuator: '>>'  op: C99RightShift].
						  (self skip: '>='  ) ifTrue: [^self punctuator: '>='  op: C99GreaterEqual].
						  ^self next;			     punctuator: '>'   op: C99Greater]
C99Scanner xAmpersand	" && &= & "		[ (self skip: '&&'  ) ifTrue: [^self punctuator: '&&'  op: C99LogAnd].
						  (self skip: '&='  ) ifTrue: [^self punctuator: '&='  op: C99AndInto].
						  ^self next;			     punctuator: '&'   op: C99And]
C99Scanner xBar		" || |= | "		[ (self skip: '||'  ) ifTrue: [^self punctuator: '||'  op: C99LogOr].
						  (self skip: '|='  ) ifTrue: [^self punctuator: '|='  op: C99OrInto].
						  ^self next;			     punctuator: '|'   op: C99Or]
C99Scanner xStar	" *= * "		[ (self skip: '*='  ) ifTrue: [^self punctuator: '*='  op: C99MultiplyInto].
						  ^self next;			     punctuator: '*'   op: C99Multiply]
C99Scanner xMinus	" -- -= -> - "		[ (self skip: '--'  ) ifTrue: [^self punctuator: '--'  op: C99Decrement].
						  (self skip: '-='  ) ifTrue: [^self punctuator: '-='  op: C99SubtractInto].
						  (self skip: '->'  ) ifTrue: [^self punctuator: '->'  op: C99MemberRef].
						  ^self next;			     punctuator: '-'   op: C99Subtract]
C99Scanner xPlus	" ++ += + "		[ (self skip: '++'  ) ifTrue: [^self punctuator: '++'  op: C99Increment].
						  (self skip: '+='  ) ifTrue: [^self punctuator: '+='  op: C99AddInto].
						  ^self next;			     punctuator: '+'   op: C99Add]
C99Scanner xPling	" != ! "		[ (self skip: '!='  ) ifTrue: [^self punctuator: '!='  op: C99NotEqual].
						  ^self next;			     punctuator: '!'   op: C99Not]
C99Scanner xColon	" :> : "		[ (self skip: ':>'  ) ifTrue: [^self punctuator: ']'   op: C99CloseBracket].
						  ^self next;			     punctuator: ':'   op: C99Colon]
C99Scanner xEqual	" == = "		[ (self skip: '=='  ) ifTrue: [^self punctuator: '=='  op: C99Equal].
						  ^self next;			     punctuator: '='   op: C99Assign]
C99Scanner xPercent	" %:%: %: %> %= % "	[ (self skip: '%:%:') ifTrue: [^self punctuator: '##'  op: C99Paste].
						  (self skip: '%:'  ) ifTrue: [^self punctuator: '#'   op: C99Hash].
						  (self skip: '%>'  ) ifTrue: [^self punctuator: '}'   op: C99CloseBrace].
						  (self skip: '%='  ) ifTrue: [^self punctuator: '%='  op: C99ModulusInto].
						  ^self next;			     punctuator: '%'   op: C99Modulus]
C99Scanner xHat		" ^= ^ "		[ (self skip: '^='  ) ifTrue: [^self punctuator: '^='  op: C99XorInto].
						  ^self next;			     punctuator: '^'   op: C99Xor]
C99Scanner xHash	" ## # "		[ (self skip: '##'  ) ifTrue: [^self paste:      '##'  op: C99Paste or: C99Other].
						   bol		      ifTrue: [^self directive             ].
						  ^self next;			     punctuator: '#'   op: C99Hash]

C99Scanner paste: tokenText op: dirType or: textType
[
    ^(inDirective ifTrue: [dirType] ifFalse: [textType])
	withText: tokenText file: tokenFileName line: tokenLineNumber
]

" directives "

C99Scanner directive
[
    | start token dir |
    bol ifFalse: [self error: 'directive processing reentered after punt'].
    bol := false.
    start := line position.
    self next.
    [(token := self nextToken) isBlank and: [token notEOL]] whileTrue.
    (token isIdentifier and: [(dir := C99DirectiveTable at: token text ifAbsent: []) notNil])
	ifFalse:
	   [line position: start.	"tokenize again ignoring directives"
	    ^self nextToken].
    inDirective := true.
    (dir := dir withText: token text file: tokenFileName line: tokenLineNumber)
	name:		 token;
	scanFrom:	 self;
	scanDirectiveIn: self.
    inDirective := false.
    ^dir
]

C99Directive scanFrom: scanner
[
    | tok |
    [(tok := scanner nextToken) notNil and: [tok notEOL]] whileTrue: [body add: tok].
]

C99DirectiveIf	    scanDirectiveIn: scanner	[ ^scanner scanIf:	self ]	"inherited by dirIfdef dirIfndef"
C99DirectiveElif    scanDirectiveIn: scanner	[ ^scanner scanElif:    self ]
C99DirectiveElse    scanDirectiveIn: scanner	[ ^scanner scanElse:    self ]
C99DirectiveEndif   scanDirectiveIn: scanner	[ ^scanner scanEndif:   self ]
" no-ops during scan: "
C99DirectiveInclude scanDirectiveIn: scanner	[]
C99DirectiveDefine  scanDirectiveIn: scanner	[]
C99DirectiveUndef   scanDirectiveIn: scanner	[]
C99DirectiveLine    scanDirectiveIn: scanner	[]
C99DirectiveError   scanDirectiveIn: scanner	[]
C99DirectivePragma  scanDirectiveIn: scanner	[]

" build group structure within conditional sections "

C99Scanner scanIf: token
[
    token enclosingGroup: activeGroup.
    activeGroup := token.
]

C99Scanner scanElif: token
[
    activeGroup isNil ifTrue: [token error: '#elif without #if'].
    activeGroup isDirectiveElse ifTrue: [token error: '#elif follows #else'].
    activeGroup nextGroup: token.
    token enclosingGroup: activeGroup enclosingGroup.
    activeGroup := token.
]

C99Scanner scanElse: token
[
    activeGroup isNil ifTrue: [token error: '#else without #if'].
    activeGroup isDirectiveElse ifTrue: [token error: '#else follows #else'].
    activeGroup nextGroup: token.
    token enclosingGroup: activeGroup enclosingGroup.
    activeGroup := token.
]

C99Scanner scanEndif: token
[
    activeGroup isNil ifTrue: [token error: '#endif without #if'].
    activeGroup nextGroup: token.
    activeGroup := activeGroup enclosingGroup.
]

"----------------------------------------------------------------"

" a scanner that ignores all directives (used by ## operator) "

C99Rescanner : C99Scanner ()

C99Rescanner beginningOfLine	[ bol := false ]
