#ifndef __vpu_VPU_h
#define __vpu_VPU_h

#define VPU_MAJOR	4
#define VPU_MINOR	0
#define VPU_PATCH	3

class Addr;
class AddrList;
class Arg;
class ArgStack;
class CPU;
class Entry;
class FlowGraph;
class Heap;
class Insn;
class Label;
class LabelList;
class LabelStack;
class LocalLabel;
class RegSet;
class Stack;
class Temp;
class TempStack;
class VarList;

class VPU
{
public:
  typedef void *(*Malloc)(unsigned int);
  typedef void  (*Free)(void *);

  static Malloc malloc;
  static Free   free;

  static char *versionString;

  unsigned flags;	// mask of the following...
  enum {
    // client debugging options:
    DumpAsm= 0,		// dump final assembly listing
    DumpCode,		// dump final pseudo-code before assembly
    // optimisation options:
    OptFold,		// fold constant expressions (cheap)
    OptMove,		// eliminate stubborn move chains (uncommon)
    // code generation options:
    GenLongCalls,	// support arbitrary (32-bit) call/jump destinations
  };
  static unsigned debug;
  enum {
    // internal debugging otions (unsupported in production vpu):
    DebugStack= 0,	// dump initial stack
    DebugFlow,		// dump flow analysis
    DebugType,		// dump type analysis
    DebugMode,		// dump initial mode selection
    DebugConstrain,	// dump register constraints
    DebugTarget,	// dump register targets
    DebugReg,		// dump register allocation and spill
    DebugSelect,	// dump insn selection and reload
    DebugFrame,		// dump frame initialisation
    DebugSize,		// dump first assembly pass (code sizing)
    DebugRelocate,	// dump label relocation
    DebugOpt,		// dump optimisation passes
    DebugVerbose,	// be (much) more verbose when dumping any of the above
    DebugColour,	// use ANSI colour sequences to improve readability
  };
  Heap       *heap;
  Insn       *prog;
  Insn       *last;
  Stack	     *stack;
  ArgStack   *args;
  TempStack  *temps;
  AddrList   *addrs;
  Entry	     *entry;
  LabelStack *labels;
  LabelList  *relocations;
  FlowGraph  *flow;
  CPU	     *cpu;
  RegSet    **regs;
  VarList   **vars;
  unsigned    avail[2];
  bool	      leaf;
  bool	      variadic;
  bool	      dynamic;

   VPU();
  ~VPU(void);

  Insn *addLast(Insn *insn);

  Arg  *pushArgs(Arg *insn);
  Insn *pushStack(Insn *insn);
  VPU  *popStack(int n= 1);
  Temp *pushTemps(Temp *insn);
  VPU  *popTemps(int n= 1);

  void noteFlow(Insn *src, Insn *dst);
  int  resolveFlow(Insn *src, Insn *dst);
  void resolveFlow(void);

  // labels

#if (VPU_MAJOR == 1)
# include "compat.h"
#else

  LocalLabel *labelAt(int n);
  LocalLabel *newLabel(void);
  void       *labelValue(LocalLabel *l);

  VPU	*Begin(int n= 1);
  VPU	*Define(int n);
  VPU	*Define(Label &l);
  VPU	*Define(Label *l);
  VPU	*Define(LocalLabel *l);
  VPU	*End(int n= 1);

  // function, args, temps

  VPU	*Ienter(void),		*Lenter(void);
  VPU	*Iarg(int n= 1),	*Larg(int n= 1);
  VPU	*Itmp(int n= 1),	*Ltmp(int n= 1);
  VPU	*LdArg(int n);
  VPU	*StArg(int n);
  VPU	*LdTmp(int n);
  VPU	*StTmp(int n);
  VPU	*DropTmp(int n= 1);
  VPU	*LdFP(void);
  VPU	*AddrTmp(int n);
  VPU	*AddrArg(int n);
  VPU	*Addr(LocalLabel *l);
  VPU	*SaveArgs(void);
  VPU	*Alloc(void);

  // stack

  VPU	*Ld(int i);
  VPU	*Ld(long long i);
  VPU	*Ld(char *s);
  VPU	*Ld(void *p);
  VPU	*Dup(int n= 0);
  VPU	*Put(int n);
  VPU	*Drop(int n= 1);

  // conditional

  VPU	*Lt(void);
  VPU	*Le(void);
  VPU	*Eq(void);
  VPU	*Ne(void);
  VPU	*Ge(void);
  VPU	*Gt(void);
  VPU	*Not(void);

  VPU	*Bf(int n);
  VPU	*Bt(int n);
  VPU	*Br(int n);
  VPU	*Br(LocalLabel *l);
  VPU	*Br(void);

  // call

  VPU	*Icall(int n);
  VPU	*Icall(int n, void *fn);
  VPU	*Icall(int n, Label *l);
  VPU	*Ret(void);

  // arithmetic

  VPU	*Icvt(void);

  VPU	*Neg(void);

  VPU	*Add(void);
  VPU	*Sub(void);
  VPU	*Div(void);
  VPU	*Mul(void);
  VPU	*Mod(void);
  VPU	*And(void);
  VPU	*Or (void);
  VPU	*Xor(void);
  VPU	*Asl(void);
  VPU	*Asr(void);
  VPU	*Lsl(void);
  VPU	*Lsr(void);

  // memory

  VPU	*Rdb(void);
  VPU	*Rdh(void);
  VPU	*Rdw(void);
  VPU	*Wrb(void);
  VPU	*Wrh(void);
  VPU	*Wrw(void);

# endif // VPU_MAJOR != 1

  // compilation

  VPU  *dump(void);
  void *compile(unsigned flags= 0, unsigned debug= 0);
  void *codeStart(void);
  int   codeSize(void);
};


#endif // __vpu_VPU_h
