#if defined(__MACH__) || defined(__unix__)
# define CWINDOW_X11 1
#else
# error platform not supported
#endif

#include "surface-imp.h"

#if CWINDOW_X11

# include <stdlib.h>
# include <stdio.h>
# include <X11/Xlib.h>
# include <X11/Xutil.h>

struct _CWindowDevice
{
  Display	*dpy;
  Window	 win;
  Visual	*vis;
  GC		 gc;
};

struct _CPixmapDevice
{
  XImage	*image;
  unsigned long	 (*getPixel)(CPixmap *pix, int x, int y);
  int		 freeBits;
  int		 bytesPerLine;
};

static void fatal(char *message)
{
  fprintf(stderr, "%s\n", message);
  exit(1);
}

static CWindow *newCWindowDevice(CWindow *win)
{
  CWindowDevice *dev= (CWindowDevice *)calloc(1, sizeof(CWindowDevice));
  dev->dpy= XOpenDisplay(0);
  win->depth= DefaultDepth(dev->dpy, DefaultScreen(dev->dpy));
  if (!dev->dpy) fatal("could not open display");
  {
    XVisualInfo xvi;
    if (!XMatchVisualInfo(dev->dpy, DefaultScreen(dev->dpy), win->depth, TrueColor, &xvi)) fatal("could not find TrueColor visual");
    dev->vis= xvi.visual;
  }
  {
    XSetWindowAttributes xwa;
    xwa.background_pixel= 0x000000;
    xwa.backing_store=	  Always;
    xwa.event_mask=	  KeyPressMask | KeyReleaseMask | ButtonPressMask | ButtonReleaseMask | PointerMotionMask;
    dev->win= XCreateWindow(dev->dpy, DefaultRootWindow(dev->dpy),
			    100, 100, win->width, win->height, 0, win->depth,
			    InputOutput, dev->vis,
			    CWBackPixel | CWBackingStore | CWEventMask, &xwa);
  }
  {
    XGCValues gcv;
    gcv.function=   	    GXcopy;
    gcv.plane_mask= 	    0xffffffffUL;
    gcv.foreground= 	    0xffffffUL;
    gcv.background= 	    0x000000UL;
    gcv.line_width= 	    0;
    gcv.line_style= 	    LineSolid;
    gcv.cap_style=  	    CapButt;
    gcv.join_style= 	    JoinMiter;
    gcv.fill_style= 	    FillSolid;
    gcv.fill_rule=  	    EvenOddRule;
    gcv.subwindow_mode=     ClipByChildren;
    gcv.graphics_exposures= 1;
    gcv.clip_x_origin=      0;
    gcv.clip_y_origin=      0;
    gcv.clip_mask=          None;
    dev->gc= XCreateGC(dev->dpy, dev->win,
		       GCFunction | GCPlaneMask | GCForeground | GCBackground
		       | GCLineWidth | GCLineStyle | GCCapStyle | GCJoinStyle | GCFillStyle | GCFillRule
		       | GCSubwindowMode | GCGraphicsExposures | GCClipXOrigin | GCClipYOrigin | GCClipMask,
		       &gcv);
  }
  XMapWindow(dev->dpy, dev->win);
  win->device= dev;
  return win;
}

static void deleteCWindowDevice(CWindow *win)
{
  CWindowDevice *dev= win->device;
  XFreeGC(dev->dpy, dev->gc);
  XDestroyWindow(dev->dpy, dev->win);
  XCloseDisplay(dev->dpy);
  free(dev);
}

static unsigned long getPixel1(CPixmap *pix, int x, int y)
{
  unsigned char *addr= 0;
  unsigned pixel= 0;
  addr= (unsigned char *)pix->bits + (pix->device->bytesPerLine * y) + (x / 8);
  pixel= *addr;
  return (pixel & (0x80 >> (x % 8))) ? 0x000000 : 0xFFFFFF;
}

static unsigned long getPixel16(CPixmap *pix, int x, int y)
{
  char *addr= 0;
  unsigned short pixel= 0;
  int r= 0, g= 0, b= 0;
  addr= (char *)pix->bits + (pix->device->bytesPerLine * y) + (sizeof(short) * x);
  pixel= *(unsigned short *)addr;
  r= (pixel >> 10) & 31;  r= (r << 3) + (r & 7);  r <<= 16;
  g= (pixel >>  5) & 31;  g= (g << 3) + (g & 7);  g <<=  8;
  b= (pixel >>  0) & 31;  b= (b << 3) + (b & 7);  b <<=  0;
  return r | g | b;
}

CPixmap *newCPixmapDevice(CWindow *win, CPixmap *pix)
{
  CPixmapDevice *dev= (CPixmapDevice *)calloc(1, sizeof(CPixmapDevice));
  char *data= 0;
  int pixBitsPerPixel= ((pix->depth == 24) ? 32 : pix->depth);
  int pixBytesPerLine= ((pix->width * pixBitsPerPixel + 31) & ~31) / 8;
  int winBitsPerPixel= ((win->depth == 24) ? 32 : win->depth);
  int winBytesPerLine= ((win->width * winBitsPerPixel + 31) & ~31) / 8;
  dev->bytesPerLine= pixBytesPerLine;
  if (pixBitsPerPixel == winBitsPerPixel)
    {
      data= (char *)pix->bits;
      dev->freeBits= 0;
      dev->getPixel= 0;
    }
  else
    {
      data= (char *)malloc(winBytesPerLine * pix->height);
      dev->freeBits= 1;
      switch (pix->depth)
	{
	case  1:  dev->getPixel= getPixel1;   break;
	case 16:  dev->getPixel= getPixel16;  break;
	default:
	  fatal("unsupported pixmap depth");
	  break;
	}
    }
  dev->image= XCreateImage(win->device->dpy, win->device->vis, win->depth, ZPixmap, 0, data, pix->width, pix->height, 32, winBytesPerLine);
  pix->device= dev;
  return pix;
}

void deleteCPixmapDevice(CPixmap *pix)
{
  if (!pix->device->freeBits)
    pix->device->image->data= 0;
  XDestroyImage(pix->device->image);
  free(pix->device);
}

void blitCWindowDevice(CWindow *win, int dstX, int dstY, CPixmap *pix, int srcX, int srcY, int w, int h)
{
  if (pix->device->getPixel)
    {
      int x, x2= srcX + w, y, y2= srcY + h;
      for (y= srcY;  y < y2;  ++y)
	for (x= srcX;  x < x2;  ++x)
	  XPutPixel(pix->device->image, x, y, pix->device->getPixel(pix, x, y));
    }
  XPutImage(win->device->dpy, win->device->win, win->device->gc, pix->device->image, srcX, srcY, dstX, dstY, w, h);
}

static int xKeycodeToAscii(Display *dpy, XEvent *evt)
{
  int     index= 0;
  KeySym  keySym;
  if (evt->xkey.state & ShiftMask)
    ++index;
  keySym= XKeycodeToKeysym(dpy, evt->xkey.keycode, index);
  if ((NoSymbol == keySym) && (index & 1))
    keySym= XKeycodeToKeysym(dpy, evt->xkey.keycode, index - 1);
  return keySym & 0xFF;
}

static CWindowEvent *nextCWindowEventDevice(CWindow *win, int block)
{
  static CWindowEvent sevt;
  for (;;)
    {
      if (!XPending(win->device->dpy) && !block) return 0;
      {
	XEvent xevt;
	XNextEvent(win->device->dpy, &xevt);
	switch (xevt.type)
	  {
	  case KeyPress:
	  case KeyRelease:
	    sevt.type= ((xevt.type == KeyPress) ? CWindowKeyPress : CWindowKeyRelease);
	    sevt.key.x= xevt.xkey.x;
	    sevt.key.y= xevt.xkey.y;
	    sevt.key.key= xKeycodeToAscii(win->device->dpy, &xevt);
	    return &sevt;

	  case ButtonPress:
	  case ButtonRelease:
	    sevt.type= ((xevt.type == ButtonPress) ? CWindowButtonPress : CWindowButtonRelease);
	    sevt.button.x= xevt.xbutton.x;
	    sevt.button.y= xevt.xbutton.y;
	    sevt.button.button= xevt.xbutton.button;
	    return &sevt;

	  default:
	    /* fprintf(stderr, "ignoring event type %d\n", xevt.type); */
	    break;
	  }
      }
    }
}

#endif /* CWINDOW_X11 */


CWindow *newCWindow(int w, int h)
{
  CWindow *win= calloc(1, sizeof(CWindow));
  win->width=  w;
  win->height= h;
  newCWindowDevice(win);
  return win;
}

void deleteCWindow(CWindow *win)
{
  deleteCWindowDevice(win);
  free(win);
}

static void *allocBits(int w, int h, int d)
{
  int bitsPerPixel= ((d == 24) ? 32 : d);
  int bytesPerLine= ((w * bitsPerPixel + 31) & ~31) / 8;
  /* printf("allocBits %d\n", bytesPerLine * h); */
  return malloc(bytesPerLine * h);
}

CPixmap *newCPixmap(CWindow *win, int w, int h, int d, void *bits)
{
  CPixmap *pix= (CPixmap *)calloc(1, sizeof(CPixmap));
  pix->width=  w;
  pix->height= h;
  pix->depth=  d;
  if (bits)
    {
      pix->bits= bits;
      pix->freeBits= 0;
    }
  else
    {
      pix->bits= allocBits(w, h, d);
      pix->freeBits= 1;
    }
  newCPixmapDevice(win, pix);
  return pix;
}

void deleteCPixmap(CPixmap *pix)
{
  deleteCPixmapDevice(pix);
  if (pix->freeBits)
    free(pix->bits);
  free(pix);
}

void blitCWindow(CWindow *win, int dstX, int dstY, CPixmap *src, int srcX, int srcY, int w, int h)
{
  blitCWindowDevice(win, dstX, dstY, src, srcX, srcY, w, h);
}

CWindowEvent *nextCWindowEvent(CWindow *win, int block)
{
  return nextCWindowEventDevice(win, block);
}

#if 0

#define W	200
#define H	200
#define D	16

int main(int argc, char **argv)
{
  CWindow *win= 0;
  CPixmap *r= 0, *g= 0, *b= 0;
  int i, freeze;

  win= newCWindow(W * 3 + 20, H + 20);
  r= newCPixmap(win, W, H, D, 0);
  g= newCPixmap(win, W, H, D, 0);
  b= newCPixmap(win, W, H, D, 0);

  switch (D)
    {
    case 32:
    case 24:
      for (i= 0;  i < W * H;  ++i)
	{
	  ((int *)r->bits)[i]= 0xFF0000;
	  ((int *)g->bits)[i]= 0x00FF00;
	  ((int *)b->bits)[i]= 0x0000FF;
	}
      break;

    case 16:
      for (i= 0;  i < W * H;  ++i)
	{
	  ((short *)r->bits)[i]= 31 << 10;
	  ((short *)g->bits)[i]= 31 <<  5;
	  ((short *)b->bits)[i]= 31 <<  0;
	}
      break;

    default:
      fatal("cannot initialise pixmaps at requested depth");
      break;
    }

  printf("run\n");

  i= 0;
  freeze= 0;

  for (;;)
    {
      CWindowEvent *evt= nextCWindowEvent(win, 0);
      if (evt)
	{
	  printf("evt %d\n", evt->type);
	  if ((evt->type == CWindowKeyRelease) && (evt->key.key == 27))
	    break;
	  else if (evt->type == CWindowButtonPress)
	    freeze= 1;
	  else if (evt->type == CWindowButtonRelease)
	    freeze= 0;
	}
      else if (!freeze)
	{
	  switch (i % 3)
	    {
	    case 0:
	      blitCWindow(win, 10 + W * 0, 10, r, 0, 0, W, H);
	      blitCWindow(win, 10 + W * 1, 10, g, 0, 0, W, H);
	      blitCWindow(win, 10 + W * 2, 10, b, 0, 0, W, H);
	      break;
	    case 1:
	      blitCWindow(win, 10 + W * 0, 10, g, 0, 0, W, H);
	      blitCWindow(win, 10 + W * 1, 10, b, 0, 0, W, H);
	      blitCWindow(win, 10 + W * 2, 10, r, 0, 0, W, H);
	      break;
	    case 2:
	      blitCWindow(win, 10 + W * 0, 10, b, 0, 0, W, H);
	      blitCWindow(win, 10 + W * 1, 10, r, 0, 0, W, H);
	      blitCWindow(win, 10 + W * 2, 10, g, 0, 0, W, H);
	      break;
	    }
	  i= i + 1;
	}
    }

  printf("done\n");

  deleteCPixmap(r);
  deleteCPixmap(g);
  deleteCPixmap(b);
  deleteCWindow(win);

  return 0;
}

#endif
